#include <stdio.h>
#include <string.h>
#include <mandelbrot-numerics.h>

static inline double cabs2(complex double z) {
  return creal(z) * creal(z) + cimag(z) * cimag(z);
}

static inline int channel(float c) {
  return fminf(fmaxf(roundf(255 * c), 0), 255);
}

static inline void hsv2rgb(float h, float s, float v, float *red, float *grn, float *blu) {
  float i, f, p, q, t, r, g, b;
  if (s == 0) { r = g = b = v; } else {
    h = 6 * (h - floorf(h));
    int ii = i = floorf(h);
    f = h - i;
    p = v * (1 - s);
    q = v * (1 - (s * f));
    t = v * (1 - (s * (1 - f)));
    switch(ii) {
      case 0: r = v; g = t; b = p; break;
      case 1: r = q; g = v; b = p; break;
      case 2: r = p; g = v; b = t; break;
      case 3: r = p; g = q; b = v; break;
      case 4: r = t; g = p; b = v; break;
      default:r = v; g = p; b = q; break;
    }
  }
  *red = r;
  *grn = g;
  *blu = b;
}

static inline float mix(float a, float b, float t) {
  return a * (1 - t) + b * t;
}

static inline void plot(unsigned char *pixel, complex double c, bool incircle, double closeness, bool wrongperiod, bool edge, double distance) {
  float h = cabs(c) < 0.25 ? 0 : 2.5464790894703255 * fabs(carg(c)) + 12.5464790894703255 * cabs(c);
  float s = incircle ? closeness < 4 ? 1 : 0.5 : 0.25;
  float v = incircle ? closeness < 4 ? 0.5 : 1 : 1;
  if (wrongperiod) {
    h = 0;
    s = 0;
    v = 0.5;
  }
  if (edge) {
    v = 0;
  }
  float r, g, b;
  hsv2rgb(h, s, v, &r, &g, &b);
  float t = 1 - tanh(distance);
  pixel[0] = channel(mix(r, 1, t));
  pixel[1] = channel(mix(g, 1, t));
  pixel[2] = channel(mix(b, 1, t));
}

extern int main(int argc, char **argv) {
  (void) argc;
  (void) argv;
  int width = 2048;
  int height = 2048;
  complex double center = -0.75;
  double radius = 1.5;
  int maxsteps = 32;
  double pixelspacing = radius / (height/2.0);
  double pixelspacing2 = pixelspacing * pixelspacing;
  unsigned char *pixels = malloc(3 * width * height * sizeof(*pixels));
  complex float *ms = malloc(width * height * sizeof(*ms));
  float *rs = malloc(2 * width * height *sizeof(*rs));
  bool *wp = malloc(width * height * sizeof(*wp));
  float *ds = malloc(width * height * sizeof(*ds));
  if (! pixels || ! ms || ! rs || ! wp || !ds) {
    return 1;
  }
  #pragma omp parallel for schedule(dynamic, 1024)
  for (int k = 0; k < width * height; ++k) {
    int i = k % width;
    int j = k / width;
    complex double c = center + ((i + 0.5 - width/2.0) + I * (height/2.0 - j - 0.5)) * pixelspacing;
    double distance = 1;
    {
      complex double z = 0;
      complex double dc = 0;
      for (int n = 0; n < 1000; ++n) {
        dc = 2 * z * dc + 1;
        z = z * z + c;
        double z2 = cabs2(z);
        if (z2 > 65536) {
          distance = sqrt(z2 / cabs2(dc)) * log(z2) / pixelspacing;
          break;
        }
      }
    }
    ds[k] = distance;
  }
  for (int per = 1; per <= 5; ++per) {
    for (int pre = 1; pre <= 5; ++pre) {
      char filename[16];
      snprintf(filename, 16, "%dp%d.ppm", pre, per);
      FILE *out = fopen(filename, "wb");
      if (! out) { return 1; }
      memset(rs, 0, 2 * width * height * sizeof(*rs));
      #pragma omp parallel for schedule(dynamic, 1024)
      for (int k = 0; k < width * height; ++k) {
        int i = k % width;
        int j = k / width;
        bool naive = (j + 0.5) < (height / 2.0);
        complex double c = center + ((i + 0.5 - width/2.0) + I * (height/2.0 - j - 0.5)) * pixelspacing;
        complex double m = 0;
        bool wrong_periods = false;
        if (naive) {
          m_d_misiurewicz_naive(&m, c, pre, per, maxsteps);
          if (cabs2(m) < 0.25 * 0.25) { m = 0; }
          for (int pp = 0; pp <= pre; ++pp) {
            for (int p = 1; p <= per; ++p) {
              if (pp == pre && p == per) {
                continue;
              }
              complex double mp = 0;
              if (pp == 0) {
                m_d_nucleus(&mp, m, p, maxsteps);
              } else {
                m_d_misiurewicz_naive(&mp, m, pp, p, maxsteps);
              }
              if (cabs2(m - mp) < pixelspacing2) {
                wrong_periods = true;
              }
              if (wrong_periods) { break; }
            }
            if (wrong_periods) { break; }
          }
        } else {
          m_d_misiurewicz      (&m, c, pre, per, maxsteps);
          if (cabs2(m) < 0.25 * 0.25) { m = 0; }
          for (int p = 1; p < per; ++p) {
            complex double mp = 0;
            m_d_misiurewicz      (&mp, m, pre, p, maxsteps);
            if (cabs2(m - mp) < pixelspacing2) {
              wrong_periods = true;
            }
            if (wrong_periods) { break; }
          }
        }
        ms[k] = m;
        wp[k] = wrong_periods;
      }
      #pragma omp parallel for schedule(dynamic, 1024)
      for (int k = 0; k < width * height; ++k) {
        complex double p0 = (ms[k] - center) / pixelspacing;
        int i0 = round(creal(p0) + width/2.0 - 0.5);
        int j0 = round(height/2.0 - 0.5 - cimag(p0));
        if (! (0 <= i0 && i0 < width && 0 <= j0 && j0 < height)) { continue; }
        int k0 = width * j0 + i0;
        bool shouldcalc = false;
        #pragma omp critical
        {
          shouldcalc = rs[2 * k0 + 0] == 0;
          if (shouldcalc) {
            rs[2 * k0 + 0] = -1;
          }
        }
        if (! shouldcalc) { continue; }
        int size = ceil(sqrt(width * width + height * height));
        int r0 = 1;
        int r1 = 1;
        if (cimag(ms[k0]) >= -pixelspacing) {
          bool incircle = true;
          int rr;
          for (rr = 1; rr <= size; ++rr) {
            for (int jj = j0 - rr; jj <= j0 + rr; ++jj) {
              if (! (0 <= jj && jj < height)) { continue; }
              if (  (jj >= height/2.0)) { continue; }
              int dj = jj - j0;
              for (int ii = i0 - rr; ii <= i0 + rr; ++ii) {
                if (! (0 <= ii && ii < width)) { continue; }
                int di = ii - i0;
                if (! (di * di + dj * dj <= rr * rr)) { continue; }
                if (di * di + dj * dj < (rr - 1) * (rr - 1)) { continue; }
                int kk = width * jj + ii;
                incircle = incircle && cabs2(ms[kk] - ms[k0]) <= pixelspacing2;
                if (! incircle) { break; }
              }
              if (! incircle) { break; }
            }
            if (! incircle) { break; }
          }
          r0 = rr;
        }
        if (cimag(ms[k0]) <=  pixelspacing) {
          bool incircle = true;
          int rr;
          for (rr = 1; rr <= size; ++rr) {
            for (int jj = j0 - rr; jj <= j0 + rr; ++jj) {
              if (! (0 <= jj && jj < height)) { continue; }
              if (! (jj >= height/2.0)) { continue; }
              int dj = jj - j0;
              for (int ii = i0 - rr; ii <= i0 + rr; ++ii) {
                if (! (0 <= ii && ii < width)) { continue; }
                int di = ii - i0;
                if (! (di * di + dj * dj <= rr * rr)) { continue; }
                if (di * di + dj * dj < (rr - 1) * (rr - 1)) { continue; }
                int kk = width * jj + ii;
                incircle = incircle && cabs2(ms[kk] - ms[k0]) <= pixelspacing2;
                if (! incircle) { break; }
              }
              if (! incircle) { break; }
            }
            if (! incircle) { break; }
          }
          r1 = rr;
        }
        rs[2 * k0 + 0] = r0;
        rs[2 * k0 + 1] = r1;
      }
      #pragma omp parallel for schedule(dynamic, 1024)
      for (int k = 0; k < width * height; ++k) {
        int i = k % width;
        int j = k / width;
        complex double c = center + ((i + 0.5 - width/2.0) + I * (height/2.0 - j - 0.5)) * pixelspacing;
        bool incircle = false;
        if (! wp[k]) {
          complex double p0 = (ms[k] - center) / pixelspacing;
          int i0 = round(creal(p0) + width/2.0 - 0.5);
          int j0 = round(height/2.0 - 0.5 - cimag(p0));
          if (0 <= i0 && i0 < width && 0 <= j0 && j0 < height) {
            int k0 = width * j0 + i0;
            incircle = cabs(c - ms[k0]) <= pixelspacing * rs[2 * k0 + ! (j <= height/2.0)];
          }
        }
        bool edge = false;
        {
          for (int jj = j - 1; jj <= j + 1; ++jj) {
            if (! (0 <= jj && jj < height)) { continue; }
            for (int ii = i - 1; ii <= i + 1; ++ii) {
              if (! (0 <= ii && ii < width)) { continue; }
              if (abs(ii) == abs(jj)) { continue; }
              int kk = width * jj + ii;
              edge = edge || ! (cabs2(ms[kk] - ms[k]) <= pixelspacing2);
              if (edge) { break; }
            }
            if (edge) { break; }
          }
        }
        plot(pixels + 3 * k, ms[k], incircle, cabs(c - ms[k]) / pixelspacing, wp[k], edge, ds[k]);
      }
      fprintf(out, "P6\n%d %d\n255\n", width, height);
      fwrite(pixels, 3 * width * height, 1, out);
      fclose(out);
    }
  } 
  return 0;
}
